use std::collections::HashMap;

use crate::{
    ast::{BinaryType, Expr, Node, Parameter, UnaryType},
    bytecode::{Function, Instruction, Module},
    tokens::Position,
};

#[derive(Debug)]
pub struct CompileError {
    message: String,
    pos: Position,
}

pub struct ModuleContext {
    functions: Vec<Function>,
    global_functions: HashMap<String, Local>,
}

pub fn compile_module(ast: Node<Expr>) -> Result<Module, CompileError> {
    let mut context = ModuleContext {
        functions: vec![],
        global_functions: HashMap::new(),
    };

    context.functions.push(Function::BuiltIn { arguments: 1 });
    context.global_functions.insert(
        "print".to_string(),
        Local {
            address: 0,
            mutable: false,
        },
    );
    context.functions.push(Function::BuiltIn { arguments: 1 });
    context.global_functions.insert(
        "println".to_string(),
        Local {
            address: 1,
            mutable: false,
        },
    );

    let start_function = FunctionCompiler::new(&mut context).compile(Vec::new(), ast)?;
    let start_function_index = context.functions.len();
    context.functions.push(start_function);
    Ok(Module {
        functions: context.functions,
        entry_function: start_function_index,
    })
}

struct Local {
    address: usize,
    mutable: bool,
}

pub struct FunctionCompiler<'a> {
    instructions: Vec<Instruction>,
    context: &'a mut ModuleContext,
    locals: Vec<HashMap<String, Local>>,
    local_counter: usize,
}

impl<'a> FunctionCompiler<'a> {
    pub fn new(context: &'a mut ModuleContext) -> Self {
        Self {
            instructions: Vec::new(),
            context,
            locals: vec![HashMap::new()],
            local_counter: 0,
        }
    }

    pub fn compile(
        mut self,
        parameters: Vec<Node<Parameter>>,
        body: Node<Expr>,
    ) -> Result<Function, CompileError> {
        let mut param_stores = Vec::<Instruction>::new();
        for param in &parameters {
            match &param.value {
                Parameter::Error(_) => todo!(),
                Parameter::Id { name, mutable } => {
                    let addr = match self.define_local(name.to_string(), *mutable) {
                        Ok(addr) => addr,
                        Err(error) => {
                            return Err(CompileError {
                                message: error,
                                pos: param.pos.clone(),
                            })
                        }
                    };
                    param_stores.push(Instruction::StoreLocal(addr));
                }
                Parameter::Spread(_) => todo!(),
            }
        }
        for store in param_stores.into_iter().rev() {
            self.push(store);
        }
        self.compile_expr(body)?;
        self.push(Instruction::Return);
        Ok(Function::UserDefined {
            arguments: parameters.len(),
            locals_reserve: self.local_counter,
            instructions: self.instructions,
        })
    }

    fn compile_expr(&mut self, expr: Node<Expr>) -> Result<(), CompileError> {
        match expr.value {
            Expr::Error(message) => Err(CompileError {
                message,
                pos: expr.pos,
            }),
            Expr::Unit => todo!(),
            Expr::Id(name) => match self.get_local(&name) {
                Some(local) => {
                    self.push(Instruction::LoadLocal(local.address));
                    Ok(())
                }
                None => {
                    return Err(CompileError {
                        message: format!("id \"{}\" not defined", name),
                        pos: expr.pos,
                    })
                }
            },
            Expr::Int(value) => {
                self.push(Instruction::PushInt(value));
                Ok(())
            }
            Expr::Float(value) => {
                self.push(Instruction::PushFloat(value));
                Ok(())
            }
            Expr::String(value) => {
                self.push(Instruction::PushString(value));
                Ok(())
            }
            Expr::Bool(value) => {
                self.push(Instruction::PushBool(value));
                Ok(())
            }
            Expr::Array(_) => todo!(),
            Expr::Object(_) => todo!(),
            Expr::Tuple(_) => todo!(),
            Expr::If {
                condition,
                truthy,
                falsy,
            } => {
                self.compile_expr(*condition)?;
                self.push(Instruction::Not);
                let from_then_to_else = self.instructions.len();
                self.push(Instruction::JumpIf(0));
                self.compile_expr(*truthy)?;
                if let Some(expr) = falsy {
                    self.instructions[from_then_to_else] =
                        Instruction::JumpIf(self.instructions.len() + 1);
                    let from_else_to_end = self.instructions.len();
                    self.push(Instruction::Jump(0));
                    self.compile_expr(*expr)?;
                    self.instructions[from_else_to_end] =
                        Instruction::Jump(self.instructions.len());
                } else {
                    self.instructions[from_then_to_else] =
                        Instruction::JumpIf(self.instructions.len());
                }
                Ok(())
            }
            Expr::FunctionValue { parameters, body } => todo!(),
            Expr::Member { subject, value } => todo!(),
            Expr::Index { subject, value } => todo!(),
            Expr::Call { subject, arguments } => {
                for arg in arguments {
                    self.compile_expr(arg)?;
                }
                self.compile_expr(*subject)?;
                self.push(Instruction::Call);
                Ok(())
            }
            Expr::Unary {
                unary_type,
                subject,
            } => {
                self.compile_expr(*subject)?;
                self.push(match unary_type {
                    UnaryType::Not => Instruction::Not,
                    UnaryType::Negate => Instruction::Negate,
                    UnaryType::Reference => todo!(),
                    UnaryType::ReferenceMut => todo!(),
                    UnaryType::Dereference => todo!(),
                });
                Ok(())
            }
            Expr::Binary {
                binary_type,
                left,
                right,
            } => {
                self.compile_expr(*left)?;
                self.compile_expr(*right)?;
                self.push(match binary_type {
                    BinaryType::Exponentiate => Instruction::Exponentiate,
                    BinaryType::Multiply => Instruction::Multiply,
                    BinaryType::Divide => Instruction::Divide,
                    BinaryType::Modulo => Instruction::Modulo,
                    BinaryType::Add => Instruction::Add,
                    BinaryType::Subtract => Instruction::Subtract,
                    BinaryType::LT => Instruction::LT,
                    BinaryType::LTE => Instruction::LTE,
                    BinaryType::GT => Instruction::GT,
                    BinaryType::GTE => Instruction::GTE,
                    BinaryType::In => Instruction::In,
                    BinaryType::NotIn => Instruction::NotIn,
                    BinaryType::Equal => Instruction::Equal,
                    BinaryType::Inequal => Instruction::Inequal,
                    BinaryType::And => Instruction::And,
                    BinaryType::Or => Instruction::Or,
                });
                Ok(())
            }
            Expr::RangeExclusive { begin, end } => todo!(),
            Expr::RangeInclusive { begin, end } => todo!(),
            Expr::Assign {
                assign_type,
                subject,
                value,
            } => todo!(),
            Expr::Let { subject, value } => match subject.value {
                Parameter::Error(_) => todo!(),
                Parameter::Id { name, mutable } => {
                    let addr = match self.define_local(name, mutable) {
                        Ok(addr) => addr,
                        Err(error) => {
                            return Err(CompileError {
                                message: error,
                                pos: expr.pos,
                            })
                        }
                    };
                    match value {
                        Some(expr) => {
                            self.compile_expr(*expr)?;
                            self.push(Instruction::StoreLocal(addr));
                        }
                        None => {}
                    }
                    self.push(Instruction::PushUnit);
                    Ok(())
                }
                Parameter::Spread(_) => todo!(),
            },
            Expr::Continue => todo!(),
            Expr::Break => todo!(),
            Expr::While { condition, body } => todo!(),
            Expr::For {
                subject,
                value,
                body,
            } => todo!(),
            Expr::Return(value) => {
                if let Some(value) = value {
                    self.compile_expr(*value)?;
                } else {
                    self.push(Instruction::PushUnit);
                }
                self.push(Instruction::Return);
                Ok(())
            }
            Expr::Function {
                name,
                parameters,
                body,
            } => {
                let function_compiler = FunctionCompiler::new(self.context);
                let function = function_compiler.compile(parameters, *body)?;
                let function_addr = self.context.functions.len();
                self.context.functions.push(function);
                let local_addr = match self.define_local(name, false) {
                    Ok(addr) => addr,
                    Err(message) => {
                        return Err(CompileError {
                            message,
                            pos: expr.pos,
                        })
                    }
                };
                self.push(Instruction::PushFunctionReference(function_addr));
                self.push(Instruction::StoreLocal(local_addr));
                self.push(Instruction::PushUnit);
                Ok(())
            }
            Expr::Block { statements, value } => {
                self.enter_scope();
                for s in statements {
                    self.compile_expr(s)?;
                    self.instructions.pop().expect("pushed");
                }
                if let Some(value) = value {
                    self.compile_expr(*value)?;
                } else {
                    self.push(Instruction::PushUnit);
                };
                self.leave_scope();
                Ok(())
            }
            Expr::Spread(_) => todo!(),
        }
    }

    fn enter_scope(&mut self) {
        self.locals.push(HashMap::new())
    }

    fn leave_scope(&mut self) {
        self.locals.pop().unwrap();
    }

    fn define_local(&mut self, name: String, mutable: bool) -> Result<usize, String> {
        if self.locals.last().unwrap().contains_key(&name) {
            Err(format!("identifier \"{}\" already defined", name))
        } else {
            let address = self.local_counter;
            self.local_counter += 1;
            self.locals
                .last_mut()
                .unwrap()
                .insert(name, Local { address, mutable });
            Ok(address)
        }
    }

    fn get_local(&self, name: &str) -> Option<&Local> {
        for table in self.locals.iter().rev() {
            if table.contains_key(name) {
                return table.get(name);
            }
        }
        if self.context.global_functions.contains_key(name) {
            if let Some(local) = self.context.global_functions.get(name) {
                return Some(local);
            }
        }
        None
    }

    fn push(&mut self, instruction: Instruction) {
        self.instructions.push(instruction);
    }
}
