use std::collections::VecDeque;

use crate::{
    builtins::{builtin_print, builtin_println},
    bytecode::{Function, Instruction, Module},
};

#[derive(Debug, Clone)]
pub enum Value {
    Unit,
    Int(i64),
    Float(f64),
    String(String),
    Bool(bool),
    FunctionReference(usize),
    HeapReference(usize, usize),
}

#[derive(Debug)]
pub struct RuntimeError {
    pub message: String,
}

struct Call {
    return_addr: usize,
    instruction_ptr: usize,
    stack_base: usize,
    locals_base: usize,
}

struct HeapEntry {
    value: Value,
    generation: usize,
    gc_marker: bool,
}

pub struct Runtime {
    functions: Vec<Function>,
    call_stack: Vec<Call>,
    stack: Vec<Value>,
    stack_base: usize,
    locals: Vec<Value>,
    locals_base: usize,
    heap: Vec<Option<HeapEntry>>,
    heap_generation: usize,
    deallocated_addrs: VecDeque<usize>,
    function_ptr: usize,
    instruction_ptr: usize,
}

impl Runtime {
    pub fn new(module: Module) -> Self {
        Self {
            functions: module.functions,
            call_stack: Vec::new(),
            stack: Vec::new(),
            stack_base: 0,
            locals: Vec::new(),
            locals_base: 0,
            heap: Vec::new(),
            heap_generation: 0,
            deallocated_addrs: VecDeque::new(),
            function_ptr: module.entry_function,
            instruction_ptr: 0,
        }
    }

    pub fn run(mut self) -> Result<Value, RuntimeError> {
        self.call_stack.push(Call {
            return_addr: 0,
            instruction_ptr: 0,
            stack_base: 0,
            locals_base: 0,
        });
        for _ in 0..self.functions[self.function_ptr].locals_reserve() {
            self.locals.push(Value::Unit);
        }
        while self.call_stack.len() > 0 {
            let debug = false;
            if debug {
                println!("function_ptr = {}", self.function_ptr);
                println!("instruction_ptr = {}", self.instruction_ptr);
                println!("ins = {:?}", self.current_instruction());
                println!("stack before = {:?}", self.stack);
                println!("locals before = {:?}", self.locals);
            }
            self.run_instruction()?;
            if debug {
                println!("stack after = {:?}", self.stack);
                println!("locals after = {:?}", self.locals);
                println!("\n");
            }
        }
        self.stack
            .pop()
            .ok_or_else(|| self.error_value("stack empty"))
    }

    fn run_instruction(&mut self) -> Result<(), RuntimeError> {
        if self.instruction_ptr >= self.current_function().instructions().len() {
            return self.error("exceeded function end");
        }
        match &self.functions[self.function_ptr].instructions()[self.instruction_ptr] {
            Instruction::Duplicate => {
                self.assert_stack_amount(1)?;
                self.stack
                    .push(self.stack.last().expect("stack size asserted").clone());
                self.step_and_ok()
            }
            Instruction::Pop => {
                self.assert_stack_amount(1)?;
                if self.stack.len() == self.stack_base {}
                let _ = self.stack.pop().expect("stack size asserted");
                self.step_and_ok()
            }
            Instruction::Jump(addr) => {
                if !(0..self.current_function().instructions().len()).contains(&addr) {
                    return self.error("address outside function");
                }
                self.instruction_ptr = *addr;
                Ok(())
            }
            Instruction::JumpIf(addr) => {
                self.assert_stack_amount(1)?;
                let value = self.stack.pop().expect("stack size asserted");
                let should_jump = match value {
                    Value::Bool(should_jump) => should_jump,
                    _ => return self.error("expected bool value"),
                };
                if !(0..self.current_function().instructions().len()).contains(&addr) {
                    return self.error("instruction address out of range");
                }
                if should_jump {
                    self.instruction_ptr = *addr;
                } else {
                    self.step();
                }
                Ok(())
            }
            Instruction::Call => {
                self.assert_stack_amount(1)?;
                let value = self.stack.pop().expect("stack size asserted");
                let addr = match value {
                    Value::FunctionReference(addr) => addr,
                    _ => return self.error("expected function reference"),
                };
                if !(0..self.functions.len()).contains(&addr) {
                    return self.error("function address out of range");
                }
                let function = &self.functions[addr];
                if let Function::BuiltIn { .. } = function {
                    let value = match addr {
                        0 => builtin_print(&mut self.stack)?,
                        1 => builtin_println(&mut self.stack)?,
                        _ => return self.error("unknown builtin function"),
                    };
                    self.stack.push(value);
                    return Ok(());
                }
                if self.stack.len() < self.stack_base + function.arguments() {
                    return self.error(&format!(
                        "function expects {} arguments, {} present",
                        function.arguments(),
                        self.stack.len() - self.stack_base
                    ));
                }
                self.call_stack.push(Call {
                    return_addr: self.function_ptr,
                    instruction_ptr: self.instruction_ptr + 1,
                    stack_base: self.stack_base,
                    locals_base: self.locals_base,
                });
                self.stack_base = self.stack.len() - function.arguments();
                self.locals_base = self.locals.len();
                self.locals
                    .reserve(self.locals_base + function.locals_reserve());
                for _ in 0..function.locals_reserve() {
                    self.locals.push(Value::Unit);
                }
                self.function_ptr = addr;
                self.instruction_ptr = 0;
                Ok(())
            }
            Instruction::Return => {
                let call = self
                    .call_stack
                    .pop()
                    .ok_or_else(|| self.error_value("call stack underflow"))?;
                if self.stack.len() != self.stack_base + 1 {
                    return self.error(&format!(
                        "expected 1 value on stack on return, got {}",
                        self.stack.len() - self.stack_base
                    ));
                }
                self.stack_base = call.stack_base;
                for _ in 0..self.locals.len() - self.locals_base {
                    self.locals.pop().expect("size asserted");
                }
                self.locals_base = call.locals_base;
                self.function_ptr = call.return_addr;
                self.instruction_ptr = call.instruction_ptr;
                Ok(())
            }
            Instruction::LoadLocal(addr) => {
                if *addr == 0 {
                    self.stack.push(Value::FunctionReference(0));
                    return self.step_and_ok();
                } else if *addr == 1 {
                    self.stack.push(Value::FunctionReference(1));
                    return self.step_and_ok();
                }
                if !(0..self.locals.len() - self.locals_base).contains(&addr) {
                    return self.error("local address out of range");
                }
                self.stack
                    .push(self.locals[self.locals_base + addr].clone());
                self.step_and_ok()
            }
            Instruction::StoreLocal(addr) => {
                self.assert_stack_amount(1)?;
                if !(0..self.locals.len() - self.locals_base).contains(&addr) {
                    return self.error("local address out of range");
                }
                self.locals[self.locals_base + *addr] = self.stack.pop().expect("value");
                self.step_and_ok()
            }
            Instruction::HeapLoad => {
                self.assert_stack_amount(1)?;
                let (addr, generation) = match self.stack.pop() {
                    Some(Value::HeapReference(addr, generation)) => (addr, generation),
                    _ => return self.error("expected heap address"),
                };
                if !(0..self.heap.len()).contains(&addr) {
                    return self.error("heap address out of range");
                };
                let entry = match &self.heap[addr] {
                    Some(entry) if entry.generation == generation => entry,
                    _ => return self.error("invalid heap address"),
                };
                self.stack.push(entry.value.clone());
                self.step_and_ok()
            }
            Instruction::HeapStore => {
                self.assert_stack_amount(2)?;
                let (addr, generation) = match self.stack.pop() {
                    Some(Value::HeapReference(addr, generation)) => (addr, generation),
                    _ => return self.error("expected heap address"),
                };
                if !(0..self.heap.len()).contains(&addr) {
                    return self.error("heap address out of range");
                };
                let entry = match &mut self.heap[addr] {
                    Some(entry) if entry.generation == generation => entry,
                    _ => return self.error("invalid heap address"),
                };
                entry.value = self.stack.pop().expect("stack size asserted");
                Ok(())
            }
            Instruction::HeapAlloc => {
                let addr = if let Some(addr) = self.deallocated_addrs.pop_front() {
                    self.heap[addr] = Some(HeapEntry {
                        value: Value::Unit,
                        generation: self.heap_generation,
                        gc_marker: false,
                    });
                    addr
                } else {
                    let addr = self.heap.len();
                    self.heap.push(Some(HeapEntry {
                        value: Value::Unit,
                        generation: self.heap_generation,
                        gc_marker: false,
                    }));
                    addr
                };
                self.stack
                    .push(Value::HeapReference(addr, self.heap_generation));
                self.heap_generation = self.heap_generation.wrapping_add(1);
                self.step_and_ok()
            }
            Instruction::PushUnit => {
                self.stack.push(Value::Unit);
                self.step_and_ok()
            }
            Instruction::PushInt(value) => {
                self.stack.push(Value::Int(*value));
                self.step_and_ok()
            }
            Instruction::PushFloat(value) => {
                self.stack.push(Value::Float(*value));
                self.step_and_ok()
            }
            Instruction::PushString(value) => {
                self.stack.push(Value::String(value.clone()));
                self.step_and_ok()
            }
            Instruction::PushBool(value) => {
                self.stack.push(Value::Bool(*value));
                self.step_and_ok()
            }
            Instruction::PushFunctionReference(value) => {
                self.stack.push(Value::FunctionReference(*value));
                self.step_and_ok()
            }
            Instruction::Not => {
                self.assert_stack_amount(1)?;
                match self.stack.pop().expect("stack size asserted") {
                    Value::Bool(value) => {
                        self.stack.push(Value::Bool(!value));
                        self.step_and_ok()
                    }
                    _ => self.error("invalid type for 'not', expected bool"),
                }
            }
            Instruction::Negate => {
                self.assert_stack_amount(1)?;
                match self.stack.pop().expect("stack size asserted") {
                    Value::Int(value) => {
                        self.stack.push(Value::Int(-value));
                        self.step_and_ok()
                    }
                    Value::Float(value) => {
                        self.stack.push(Value::Float(-value));
                        self.step_and_ok()
                    }
                    _ => self.error("invalid type for 'negate', expected int or float"),
                }
            }
            Instruction::Add => {
                self.assert_stack_amount(2)?;
                match (
                    self.stack.pop().expect("stack size asserted"),
                    self.stack.pop().expect("stack size asserted"),
                ) {
                    (Value::Int(right), Value::Int(left)) => {
                        self.stack.push(Value::Int(left + right));
                        self.step_and_ok()
                    }
                    (Value::Int(right), Value::Float(left)) => {
                        self.stack.push(Value::Float(left + right as f64));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Int(left)) => {
                        self.stack.push(Value::Float((left as f64) + right));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Float(left)) => {
                        self.stack.push(Value::Float(left + right));
                        self.step_and_ok()
                    }
                    _ => self.error("invalid type for 'add', expected (int|float, int|float)"),
                }
            }
            Instruction::Subtract => {
                self.assert_stack_amount(2)?;
                match (
                    self.stack.pop().expect("stack size asserted"),
                    self.stack.pop().expect("stack size asserted"),
                ) {
                    (Value::Int(right), Value::Int(left)) => {
                        self.stack.push(Value::Int(left - right));
                        self.step_and_ok()
                    }
                    (Value::Int(right), Value::Float(left)) => {
                        self.stack.push(Value::Float(left - right as f64));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Int(left)) => {
                        self.stack.push(Value::Float((left as f64) - right));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Float(left)) => {
                        self.stack.push(Value::Float(left - right));
                        self.step_and_ok()
                    }
                    _ => self.error("invalid type for 'subtract', expected (int|float, int|float)"),
                }
            }
            Instruction::Multiply => {
                self.assert_stack_amount(2)?;
                match (
                    self.stack.pop().expect("stack size asserted"),
                    self.stack.pop().expect("stack size asserted"),
                ) {
                    (Value::Int(right), Value::Int(left)) => {
                        self.stack.push(Value::Int(left * right));
                        self.step_and_ok()
                    }
                    (Value::Int(right), Value::Float(left)) => {
                        self.stack.push(Value::Float(left * right as f64));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Int(left)) => {
                        self.stack.push(Value::Float((left as f64) * right));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Float(left)) => {
                        self.stack.push(Value::Float(left * right));
                        self.step_and_ok()
                    }
                    _ => self.error("invalid type for 'multiply', expected (int|float, int|float)"),
                }
            }
            Instruction::Divide => {
                self.assert_stack_amount(2)?;
                match (
                    self.stack.pop().expect("stack size asserted"),
                    self.stack.pop().expect("stack size asserted"),
                ) {
                    (Value::Int(right), Value::Int(left)) => {
                        self.stack.push(Value::Int(left / right));
                        self.step_and_ok()
                    }
                    (Value::Int(right), Value::Float(left)) => {
                        self.stack.push(Value::Float(left / right as f64));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Int(left)) => {
                        self.stack.push(Value::Float((left as f64) / right));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Float(left)) => {
                        self.stack.push(Value::Float(left / right));
                        self.step_and_ok()
                    }
                    _ => self.error("invalid type for 'divide', expected (int|float, int|float)"),
                }
            }
            Instruction::Modulo => {
                self.assert_stack_amount(2)?;
                match (
                    self.stack.pop().expect("stack size asserted"),
                    self.stack.pop().expect("stack size asserted"),
                ) {
                    (Value::Int(right), Value::Int(left)) => {
                        self.stack.push(Value::Int(left % right));
                        self.step_and_ok()
                    }
                    (Value::Int(right), Value::Float(left)) => {
                        self.stack.push(Value::Float(left % right as f64));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Int(left)) => {
                        self.stack.push(Value::Float((left as f64) % right));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Float(left)) => {
                        self.stack.push(Value::Float(left % right));
                        self.step_and_ok()
                    }
                    _ => self.error("invalid type for 'modulo', expected (int|float, int|float)"),
                }
            }
            Instruction::Exponentiate => {
                self.assert_stack_amount(2)?;
                match (
                    self.stack.pop().expect("stack size asserted"),
                    self.stack.pop().expect("stack size asserted"),
                ) {
                    (Value::Int(right), Value::Int(left)) => {
                        self.stack.push(Value::Int(left.pow(right as u32)));
                        self.step_and_ok()
                    }
                    (Value::Int(right), Value::Float(left)) => {
                        self.stack.push(Value::Float(left.powf(right as f64)));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Int(left)) => {
                        self.stack.push(Value::Float((left as f64).powf(right)));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Float(left)) => {
                        self.stack.push(Value::Float(left.powf(right)));
                        self.step_and_ok()
                    }
                    _ => self
                        .error("invalid type for 'exponentiate', expected (int|float, int|float)"),
                }
            }
            Instruction::LT => {
                self.assert_stack_amount(2)?;
                match (
                    self.stack.pop().expect("stack size asserted"),
                    self.stack.pop().expect("stack size asserted"),
                ) {
                    (Value::Int(right), Value::Int(left)) => {
                        self.stack.push(Value::Bool(left < right));
                        self.step_and_ok()
                    }
                    (Value::Int(right), Value::Float(left)) => {
                        self.stack.push(Value::Bool(left < right as f64));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Int(left)) => {
                        self.stack.push(Value::Bool((left as f64) < right));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Float(left)) => {
                        self.stack.push(Value::Bool(left < right));
                        self.step_and_ok()
                    }
                    _ => {
                        self.error("invalid type for 'less than', expected (int|float, int|float)")
                    }
                }
            }
            Instruction::LTE => {
                self.assert_stack_amount(2)?;
                match (
                    self.stack.pop().expect("stack size asserted"),
                    self.stack.pop().expect("stack size asserted"),
                ) {
                    (Value::Int(right), Value::Int(left)) => {
                        self.stack.push(Value::Bool(left <= right));
                        self.step_and_ok()
                    }
                    (Value::Int(right), Value::Float(left)) => {
                        self.stack.push(Value::Bool(left <= right as f64));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Int(left)) => {
                        self.stack.push(Value::Bool((left as f64) <= right));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Float(left)) => {
                        self.stack.push(Value::Bool(left <= right));
                        self.step_and_ok()
                    }
                    _ => self.error(
                        "invalid type for 'less than or equal', expected (int|float, int|float)",
                    ),
                }
            }
            Instruction::GT => {
                self.assert_stack_amount(2)?;
                match (
                    self.stack.pop().expect("stack size asserted"),
                    self.stack.pop().expect("stack size asserted"),
                ) {
                    (Value::Int(right), Value::Int(left)) => {
                        self.stack.push(Value::Bool(left > right));
                        self.step_and_ok()
                    }
                    (Value::Int(right), Value::Float(left)) => {
                        self.stack.push(Value::Bool(left > right as f64));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Int(left)) => {
                        self.stack.push(Value::Bool((left as f64) > right));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Float(left)) => {
                        self.stack.push(Value::Bool(left > right));
                        self.step_and_ok()
                    }
                    _ => self
                        .error("invalid type for 'greater than', expected (int|float, int|float)"),
                }
            }
            Instruction::GTE => {
                self.assert_stack_amount(2)?;
                match (
                    self.stack.pop().expect("stack size asserted"),
                    self.stack.pop().expect("stack size asserted"),
                ) {
                    (Value::Int(right), Value::Int(left)) => {
                        self.stack.push(Value::Bool(left >= right));
                        self.step_and_ok()
                    }
                    (Value::Int(right), Value::Float(left)) => {
                        self.stack.push(Value::Bool(left >= right as f64));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Int(left)) => {
                        self.stack.push(Value::Bool((left as f64) >= right));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Float(left)) => {
                        self.stack.push(Value::Bool(left >= right));
                        self.step_and_ok()
                    }
                    _ => self.error(
                        "invalid type for 'greater than or equal', expected (int|float, int|float)",
                    ),
                }
            }
            Instruction::Equal => {
                self.assert_stack_amount(2)?;
                match (
                    self.stack.pop().expect("stack size asserted"),
                    self.stack.pop().expect("stack size asserted"),
                ) {
                    (Value::Unit, Value::Unit) => {
                        self.stack.push(Value::Bool(true));
                        self.step_and_ok()
                    }
                    (Value::Int(right), Value::Int(left)) => {
                        self.stack.push(Value::Bool(left == right));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Float(left)) => {
                        self.stack.push(Value::Bool(left == right));
                        self.step_and_ok()
                    }
                    (Value::String(right), Value::String(left)) => {
                        self.stack.push(Value::Bool(left == right));
                        self.step_and_ok()
                    }
                    (Value::Bool(right), Value::Bool(left)) => {
                        self.stack.push(Value::Bool(left == right));
                        self.step_and_ok()
                    }
                    _ => {
                        self.error(
                                "invalid type for 'equal', expected (unit, unit), (int, int), (float, float), (string, string) or (bool, bool)"
                        )
                    }
                }
            }
            Instruction::Inequal => {
                self.assert_stack_amount(2)?;
                match (
                    self.stack.pop().expect("stack size asserted"),
                    self.stack.pop().expect("stack size asserted"),
                ) {
                    (Value::Unit, Value::Unit) => {
                        self.stack.push(Value::Bool(false));
                        self.step_and_ok()
                    }
                    (Value::Int(right), Value::Int(left)) => {
                        self.stack.push(Value::Bool(left != right));
                        self.step_and_ok()
                    }
                    (Value::Float(right), Value::Float(left)) => {
                        self.stack.push(Value::Bool(left != right));
                        self.step_and_ok()
                    }
                    (Value::String(right), Value::String(left)) => {
                        self.stack.push(Value::Bool(left != right));
                        self.step_and_ok()
                    }
                    (Value::Bool(right), Value::Bool(left)) => {
                        self.stack.push(Value::Bool(left != right));
                        self.step_and_ok()
                    }
                    _ => {
                        self.error(
                                "invalid type for 'inequal', expected (unit, unit), (int, int), (float, float), (string, string) or (bool, bool)"
                        )
                    }
                }
            }
            Instruction::And => {
                self.assert_stack_amount(2)?;
                match (
                    self.stack.pop().expect("stack size asserted"),
                    self.stack.pop().expect("stack size asserted"),
                ) {
                    (Value::Bool(right), Value::Bool(left)) => {
                        self.stack.push(Value::Bool(left && right));
                        self.step_and_ok()
                    }
                    _ => self.error("invalid type for 'and', expected (bool, bool)"),
                }
            }
            Instruction::Or => {
                self.assert_stack_amount(2)?;
                match (
                    self.stack.pop().expect("stack size asserted"),
                    self.stack.pop().expect("stack size asserted"),
                ) {
                    (Value::Bool(right), Value::Bool(left)) => {
                        self.stack.push(Value::Bool(left || right));
                        self.step_and_ok()
                    }
                    _ => self.error("invalid type for 'or', expected (bool, bool)"),
                }
            }
            Instruction::In => self.error("'not in' not implemented"),
            Instruction::NotIn => self.error("'not in' not implemented"),
        }
    }

    fn error(&self, message: &str) -> Result<(), RuntimeError> {
        Err(self.error_value(message))
    }

    fn error_value(&self, message: &str) -> RuntimeError {
        RuntimeError {
            message: message.to_string(),
        }
    }

    fn assert_stack_amount(&self, amount: usize) -> Result<(), RuntimeError> {
        if self.stack.len() < self.stack_base + amount {
            self.error(&format!(
                "expected {} value(s) on the stack, {} present",
                amount,
                self.stack.len() - self.stack_base
            ))
        } else {
            Ok(())
        }
    }

    fn step_and_ok(&mut self) -> Result<(), RuntimeError> {
        self.step();
        Ok(())
    }

    fn step(&mut self) {
        self.instruction_ptr += 1;
    }

    fn current_instruction(&self) -> &Instruction {
        &self.current_function().instructions()[self.instruction_ptr]
    }

    fn current_function(&self) -> &Function {
        &self.functions[self.function_ptr]
    }
}
